/* dupfinder v0.8 - small program to find duplicate files
 *
 * Copyright (C) 2004 Giuseppe Caulo <giusc@softzone.it>
 * based on FindDupe v1.3 by George Shearer
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <stdarg.h>
#include "strlcpy.h"
#ifndef _WIN32
# include <unistd.h>
# include <dirent.h>
#else
# include "readdir.h"
# define snprintf _snprintf
# define strcasecmp stricmp
# ifndef __STDC__
#  define stat _stat
#  define lstat _stat
#  define fstat _fstat
#  define S_IFMT   _S_IFMT
#  define S_IFDIR  _S_IFDIR
#  define S_IFCHR  _S_IFCHR
#  define S_IFREG  _S_IFREG
#  define S_IREAD  _S_IREAD
#  define S_IWRITE _S_IWRITE
#  define S_IEXEC  _S_IEXEC
# endif
typedef short nlink_t;
# define S_ISREG(x)  ((x) & _S_IFREG)
# define S_ISDIR(x)  ((x) & _S_IFDIR)
#endif
#ifndef MAX_PATH
#define MAX_PATH 256
#endif
#ifndef PATH_MAX
#define PATH_MAX MAX_PATH
#endif
#ifndef NAME_MAX
#define NAME_MAX 1
#endif
/* a few globals */
struct filelist {
   char *filename;
   off_t size;
   ino_t inode;
   nlink_t nlinks;
   char ref;
   struct filelist *next;
} *fl=(struct filelist *)0,*temp,*first=(struct filelist *)0;

int count=0,verbose;
unsigned long int wasted; /* wasted disk space */
dev_t filesys; /* used for filesystem skipping */


char *
sz_strcpy(char **dst, ...)
{
	va_list ap;
	const char *src;
	char *newdst, *dcp;
	size_t curLen, catLen, srcLen;

	if (dst == (char **) 0)
		return NULL;

	catLen = 0;
	va_start(ap, dst);
	src = va_arg(ap, char *);
	while (src != NULL) {
		catLen += strlen(src);
		src = va_arg(ap, char *);
	}
	va_end(ap);

	curLen = 0;

	newdst = malloc(curLen + catLen + 2);
	if (newdst == NULL) {
		*dst = NULL;
		return NULL;
	}

	dcp = newdst + curLen;
	va_start(ap, dst);
	src = va_arg(ap, char *);
	while (src != NULL) {
		srcLen = strlen(src);
		memcpy(dcp, src, srcLen);
		dcp += srcLen;
		src = va_arg(ap, char *);
	}
	va_end(ap);
	*dcp = '\0';

	*dst = newdst;
	return (newdst);
}	/* sz_strcpy */


int diffs(unsigned char *s1,unsigned char *s2,int x) {
   while(x-- && (unsigned char)*s1++==(unsigned char)*s2++);
   return(!(x<1&&(unsigned char)*(s1-1)==(unsigned char)*(s2-1)));
}

int diff(char *file1,char *file2) {
   int fd1=-1,fd2=-1,sr1,sr2,ret=0;
   unsigned char buf1[1024],buf2[1024];

   if(((fd1=open(file1,O_RDONLY))==-1)||((fd2=open(file2,O_RDONLY))==-1)) {
     fprintf(stderr,"Unable to compare %s & %s.\n",file1,file2);
     ret++;
   }
   else
     while(1) {
       if(((sr1=read(fd1,(unsigned char *)buf1,1024))==-1)||((sr2=read(fd2,(unsigned char *)buf2,1024))==-1)) {
         fprintf(stderr,"Read error: %s\n",(sr1==-1) ? buf1 : buf2);
         break;
       }
       if(sr1==0 && sr2==0) {
         ret=0;
         break;
       }
       if(diffs((unsigned char *)buf1,(unsigned char *)buf2,sr1)) {
         ret=1;
         break;
       }
     }
   if(fd1!=-1)
     close(fd1);
   if(fd2!=-1)
     close(fd2);
   return(ret);
}

void freeall(void) {
   fl=first;
   while(fl) {
     temp=fl->next;
     free(fl->filename);
     free(fl);
     fl=temp;
   }
}

void err_die(const char *fmt, ...) {
   va_list ap;
   va_start(ap,fmt);
   vprintf(fmt,ap);
   va_end(ap);

   freeall();
   exit(0);
}


/* malloc()'s another struct, loads the file info & increments */
void loadfile(char *pathname, const struct stat *fstats) {
   if(S_ISREG(fstats->st_mode) && ((!fstats->st_size && verbose>4) || fstats->st_size)) { /* only read regular files */
     if(!(temp=malloc(sizeof(struct filelist))))
       err_die("Out of memory");
     if (!fl) {
	   temp->next = (struct filelist *)NULL;
	 } else {
	   temp->next = fl;
	 }
	 fl = temp;
     fl->filename = NULL;
     sz_strcpy(&(fl->filename),pathname, 0);
     fl->size=fstats->st_size;
     fl->inode=fstats->st_ino;
     fl->nlinks=fstats->st_nlink;
     fl->ref=0;
     first = fl;
   }
}

int dopath(char *fullpath) {
   struct stat fstats;
   struct dirent *dirp;
   DIR *dp;
   char *ptr;

   if(lstat(fullpath,&fstats)>-1) {
     if(!S_ISDIR(fstats.st_mode)) {
       loadfile(fullpath,&fstats);
       count++;
     }
     else {
       if(!filesys)
         filesys=fstats.st_dev;
       else
         if(filesys!=fstats.st_dev) /* stay in file system */
           return(0);
       ptr=fullpath+strlen(fullpath);
       *ptr=(*(ptr-1)=='/') ? 0 : '/';
       ptr=(*ptr) ? ptr+1 : ptr;
       *ptr=0;

       if((dp=opendir(fullpath))!=NULL) {
         while((dirp=readdir(dp))!=NULL) {
           if(!strcmp(dirp->d_name,".") || !strcmp(dirp->d_name,".."))
             continue;
           strcpy(ptr,dirp->d_name);
           dopath(fullpath);
         }
         ptr[-1]=0;
         if(closedir(dp)<0)
           err_die("Can't close directory: %s\n",fullpath);
       }
     }
     return(1);
   }
   return(0);
}

int main(int argc,char *argv[]) {
   int x=0,i,j;
   int arg;
   char *fullpath=0;

   unsigned char show_help = 1;
   unsigned long min_size = 0;
   verbose = 0;
   /* Scan command Line for parameters */
    for (arg=1; arg<argc; arg++) {
	  if (*argv[arg] == '-') {
	    /* non-interactive */
	    if (*(argv[arg] + 1) == 'v') {
			if (!verbose) {
			  while (argv[arg][verbose+1] && verbose<6)
			    verbose++;
			} else {
			  err_die("You can't specify verbosity more than once\n");
			}
  	    /* help */
	    } else if (!strcmp(argv[arg], "-h") || !strcmp(argv[arg], "-?")) {
			break;
		} else if (!strcmp(argv[arg], "-min")) {
			min_size = (arg+1 < argc && *argv[arg+1] != '-') ?
			            strtoul(argv[++arg], NULL, 10) : 0;
		} else {
			err_die("Unknown option: '%s'\n", argv[arg]);
		}
	  } else {
		 show_help = 0;
		 if (!fullpath) {
           fullpath=malloc(PATH_MAX+NAME_MAX);
           if (!fullpath)
             err_die("woa!! we have terminated the memory!!!");
		 }
		 if(verbose && verbose!=3) {
		   printf("Scanning %-20s...",argv[arg]);
		   fflush(stdout);
		 }
		 j = strlen(argv[arg]);
		 strlcpy(fullpath,argv[arg], PATH_MAX);
		 if ((j > 2) && (fullpath[j - 2] != ':') && ((fullpath[j - 1] == '\\') || (fullpath[j - 1] == '/')))
		   fullpath[j - 1] = '\0';
		 x=count;
		 filesys=(dev_t)0;
		 if(!dopath(fullpath) && verbose!=3)
		   puts("Unable to lstat dir! not there?");
		 else
		   if(verbose && verbose!=3)
			 printf("loaded %d file%s.\n",count-x,(x!=1) ? "s" : "");
	  }
   }
   if(fullpath) {
     free(fullpath);
     fullpath = NULL;
   }

   if (show_help != 0)
     err_die("DupFinder v0.8\nWritten by Giuseppe Caulo.\n\n"
     		 "Usage: dupfinder [-v[v[v[v[v]]]]] [-h] [-min [minsize]] path [path ...]\n"
     		 "Where: path is the location of tree to search for duplicate files.\n");

   x=i=0;
   fl=(struct filelist *)first;

   while(fl) {
	 if (verbose == 6) {
       printf("processing -> %s\n", fl->filename);
     }
     for(temp=fl->next;temp;temp=temp->next) {
       if(temp->size==fl->size) {
         int difft;
         if ((unsigned long)temp->size < min_size)
           continue;
#ifndef _WIN32
         if(fl->inode==temp->inode || fl->ref || temp->ref) { /* exclude hard links */
           fl->ref++;
           temp->ref++;
           if (verbose == 6) {
			   printf("skipping -> %s, is an hard link to %s\n", temp->filename, fl->filename);
		   }
           continue;
         }
#endif
         x+=difft=(!diff(fl->filename,temp->filename)) ? 1 : 0;

         wasted+=(difft) ? fl->size : 0;

         if(verbose>3 || (verbose>1 && difft)) {
           if(!i++ && verbose!=3)
             printf("\nSource pathname             Target pathname               Status    Byte size\n-----------------------------------------------------------------------------\n");
           if(verbose==3) {
             if(difft)
               printf("%s %s\n",fl->filename,temp->filename);
           }
           else
             printf("%-27s %-27s (%s) (%d)\n",fl->filename,temp->filename,
                    (difft) ? "identical" : "different",(int)fl->size);
         }
       }
     }
     fl=fl->next;
   }

   freeall();

   if(verbose!=3) {
     printf("%s%d possible duplicate%s out of %d file%s.\n",(verbose>1) ? "\n" : "",x,(x!=1) ? "s" : "",count,(count!=1) ? "s" : "");
     if(x)
       printf("%lu bytes in use by duplicate files.\n",wasted);
   }
   return 0;
}

